! **********************************************************************
!
! MODULE FOR TESTING MASS CONSERVATION AND MONOTONICITY OF
! EULERIAN TRANSPORT ALGORITHMS WITH PASSIVE TRACERS
!
! MESSy SMIL
!
! Author : Patrick Joeckel, MPICH, July  2002
!
! References:
!
! **********************************************************************
#ifndef BOX
#define ECHAM5
#define MESSYCHANNELS
#endif

! **********************************************************************
MODULE messy_ptrac_si
! **********************************************************************

  ! BMIL
  USE messy_main_tools_si,   ONLY: start_message_si, end_message_si
  ! SMCL
  USE messy_main_tools,      ONLY: PTR_3D_ARRAY
  USE messy_main_tracer,     ONLY: ON, OFF                            &
                                 , SINGLE, FAMILY, ISOTOPE            &
                                 , AIR, AEROSOL, CLOUD, OCEAN, LAKE   &    
                                 , RIVER, LANDICE, SEAICE, VEGETATION &
                                 , AMOUNTFRACTION, NUMBERDENSITY      &
                                 , CONCENTRATION                      &
                                 , STRLEN_TRSET, NMAXSETID            &
                                 , MAX_CASK_R, MAX_CASK_S, MAX_CASK_I
  USE messy_main_constants_mem,  ONLY: STRLEN_MEDIUM
  USE messy_ptrac

  IMPLICIT NONE
  PRIVATE
  INTRINSIC :: NULL

  ! PUBLIC INTERFACE ROUTINES
  PUBLIC :: ptrac_initialize  ! initialize
  PUBLIC :: ptrac_init_memory ! allocate memory
  PUBLIC :: ptrac_new_tracer  ! define new tracers
  PUBLIC :: ptrac_init_tracer ! initialize tracer fields

  ! PRIVATE HELPER ROUTINES
  !PRIVATE :: ptrac_read_nml_cpl   ! initialize /CPL/-namelist )

  INTEGER, PARAMETER :: NMAXNTRAC = 100

  TYPE STR2
     CHARACTER(LEN=STRLEN_MEDIUM) :: name       = ''
     CHARACTER(LEN=STRLEN_MEDIUM) :: subname    = ''
  END TYPE STR2

  TYPE SETS
     CHARACTER(LEN=STRLEN_TRSET), DIMENSION(NMAXSETID) :: name
  END TYPE SETS

  ! GLOBAL NAMELIST SWITCHES (CPL) -------------------------------------------
  ! TRACER CHARACTERISTICS
  !   TRACER SETS
  TYPE(SETS), DIMENSION(NMAXNTRAC), SAVE :: C_SETS
  !    NAME, SUBNAME
  TYPE(STR2), DIMENSION(NMAXNTRAC), SAVE :: C_NAME
  !    UNIT
  CHARACTER(LEN=STRLEN_MEDIUM), DIMENSION(NMAXNTRAC) :: C_UNIT
  !    TYPE: SINGLE = 0, FAMILY = 1, ISOTOPE = 2
  INTEGER,   DIMENSION(NMAXNTRAC)  :: I_TYPE
  !    MEDIUM: AIR = 1, AEROSOL = 2, CLOUD = 3, OCEAN = 4, LAKE = 5, RIVER = 6
  !            LANDICE = 7, SEAICE = 8, VEGETATION = 9
  INTEGER,   DIMENSION(NMAXNTRAC)  :: I_MEDIUM
  !    QUANTITY: AMOUNTFRACTION = 1, NUMBERDENSITY = 2, CONCENTRATION = 3
  INTEGER,   DIMENSION(NMAXNTRAC)  :: I_QUANTITY
  !
  REAL(DP), DIMENSION(NMAXNTRAC, MAX_CASK_R) :: CASK_R
  CHARACTER(LEN=STRLEN_MEDIUM), DIMENSION(NMAXNTRAC, MAX_CASK_S) :: CASK_S
  INTEGER, DIMENSION(NMAXNTRAC, MAX_CASK_I) :: CASK_I

  ! ADDITIONAL AEROSOL CHARACTERISTICS
  REAL(DP),  DIMENSION(NMAXNTRAC)  :: R_AEROSOL_SIGMA
  REAL(DP),  DIMENSION(NMAXNTRAC)  :: R_AEROSOL_RADIUS
  !
  ! --------------------------------------------------------------------------

  INTEGER, SAVE  :: I_NTRAC = 0  ! ACTUAL NUMBER OF TRACERS

  ! CAHNNEL OBJECTS FOR AEROSOL
  REAL(DP), DIMENSION(:,:,:,:), POINTER :: radius_ptr => NULL()
  REAL(DP), DIMENSION(:,:,:,:), POINTER :: density_ptr => NULL()
  REAL(DP), DIMENSION(:),       POINTER :: sigma_ptr => NULL()

CONTAINS

! ------------------------------------------------------------------------
  SUBROUTINE  ptrac_initialize

    ! BMIL
    USE messy_main_mpi_si,   ONLY: p_parallel_io, p_io, p_bcast, finish
    USE messy_main_tools_si, ONLY: find_next_free_unit

    ! MESSy
    USE messy_main_tracer,   ONLY: DEFAULT_CASK_I, DEFAULT_CASK_R &
                                 , DEFAULT_CASK_S, MAX_CASK_S

    IMPLICIT NONE
    INTRINSIC :: TRIM

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'ptrac_initialize'
    INTEGER         :: iou    ! I/O unit
    INTEGER         :: status ! error status
    INTEGER         :: jt, n
    LOGICAL         :: l_no_set

    ! INIT
    ! SET DEFAULTS
    DO jt = 1, NMAXNTRAC
       C_SETS(jt)%name(:) = ''
    END DO
    ! C_NAME(:)%name = ''
    ! C_NAME(:)%subname = ''
    C_UNIT(:)                = ''
    I_TYPE(:)                = SINGLE
    I_MEDIUM(:)              = AIR
    I_QUANTITY(:)            = AMOUNTFRACTION
    !
    DO jt=1, NMAXNTRAC
       CASK_I(jt,:) = DEFAULT_CASK_I(:)
       CASK_R(jt,:) = DEFAULT_CASK_R(:)
       CASK_S(jt,:) = DEFAULT_CASK_S(:)
    END DO
    !
    R_AEROSOL_SIGMA(:)       = 0.0
    R_AEROSOL_RADIUS(:)      = 0.0

    ! INITIALIZE CPL
    IF (p_parallel_io) THEN
       iou = find_next_free_unit(100,200)
       ! *** CALL CORE ROUTINE:
       CALL ptrac_read_nml_cpl(status, iou)
       IF (status /= 0) CALL finish(substr)
       !
       ! COUNT TRACERS
       DO jt=1, NMAXNTRAC
          IF (TRIM(C_NAME(jt)%name) == '') CYCLE
          !
          l_no_set = .TRUE.
          DO n=1, NMAXSETID
             IF (TRIM(C_SETS(jt)%name(n)) /= '') THEN
                l_no_set = .FALSE.
                EXIT
             END IF
          END DO
          IF (l_no_set) CYCLE
          !
          I_NTRAC = I_NTRAC + 1
          !
          C_SETS(I_NTRAC)%name(:)        = C_SETS(jt)%name(:)
          C_NAME(I_NTRAC)%name           = C_NAME(jt)%name
          C_NAME(I_NTRAC)%subname        = C_NAME(jt)%subname
          !
          C_UNIT(I_NTRAC)                = C_UNIT(jt)                
          I_TYPE(I_NTRAC)                = I_TYPE(jt)                
          I_MEDIUM(I_NTRAC)              = I_MEDIUM(jt)              
          I_QUANTITY(I_NTRAC)            = I_QUANTITY(jt)            
          !
          CASK_I(I_NTRAC,:) = CASK_I(jt,:)
          CASK_R(I_NTRAC,:) = CASK_R(jt,:)
          CASK_S(I_NTRAC,:) = CASK_S(jt,:)
          !
          R_AEROSOL_SIGMA(I_NTRAC)       = R_AEROSOL_SIGMA(jt)       
          R_AEROSOL_RADIUS(I_NTRAC)      = R_AEROSOL_RADIUS(jt)      
          !
       END DO
    END IF
    !
    CALL p_bcast(I_NTRAC, p_io)
    !
    DO jt=1, I_NTRAC
       DO n=1, NMAXSETID
          CALL p_bcast(C_SETS(jt)%name(n), p_io)
       END DO
       CALL p_bcast(C_NAME(jt)%name, p_io)
       CALL p_bcast(C_NAME(jt)%subname, p_io)
       !
       CALL p_bcast(C_UNIT(jt), p_io)
       CALL p_bcast(I_TYPE(jt), p_io)
       CALL p_bcast(I_MEDIUM(jt), p_io)
       CALL p_bcast(I_QUANTITY(jt), p_io)
       !
       DO n=1, MAX_CASK_I
          CALL p_bcast(CASK_I(jt,n), p_io)
       END DO
       DO n=1, MAX_CASK_R
          CALL p_bcast(CASK_R(jt,n), p_io)
       END DO
       DO n=1, MAX_CASK_S
          CALL p_bcast(CASK_S(jt,n), p_io)
       END DO
       !
       CALL p_bcast(R_AEROSOL_SIGMA(jt), p_io)
       CALL p_bcast(R_AEROSOL_RADIUS(jt), p_io)
       !
    END DO

  END SUBROUTINE ptrac_initialize
! ------------------------------------------------------------------------

! ------------------------------------------------------------------------
  SUBROUTINE ptrac_new_tracer

    ! BMIL
    USE messy_main_mpi_si,        ONLY: p_parallel_io
    ! SMCL
    USE messy_main_tracer,        ONLY: new_tracer, get_tracer_set

    IMPLICIT NONE
    INTRINSIC :: TRIM

    ! LOCAL
    INTEGER :: status
    CHARACTER(LEN=*), PARAMETER :: substr = 'ptrac_new_tracer'
    INTEGER                     :: jt
    INTEGER                     :: n
    INTEGER                     :: idx

    CALL start_message_si(modstr, 'TRACER REQUEST', substr)

    tracer_loop: DO jt=1, I_NTRAC

       IF (p_parallel_io) THEN
          IF (TRIM(C_NAME(jt)%subname) ==  '') THEN
             WRITE(*,*) 'TRACER: ', TRIM(C_NAME(jt)%name)
          ELSE
             WRITE(*,*) 'TRACER: ', &
                  TRIM(C_NAME(jt)%name)//'_'//TRIM(C_NAME(jt)%subname)
          END IF
       END IF

       set_loop: DO n=1, NMAXSETID

          IF (TRIM(C_SETS(jt)%name(n)) ==  '') CYCLE

          CALL get_tracer_set(status, TRIM(C_SETS(jt)%name(n)))
          IF (status == 0) THEN
             IF (p_parallel_io) &
                  WRITE(*,*) ' ... SET '//TRIM(C_SETS(jt)%name(n))
          ELSE
             CYCLE
          END IF

          CALL new_tracer(status, C_SETS(jt)%name(n) &
               , TRIM(C_NAME(jt)%name), modstr      &
               , subname = TRIM(C_NAME(jt)%subname) &
               , idx=idx , unit=TRIM(C_UNIT(jt))    &
               , type      = I_TYPE(jt)             &
               , medium = I_MEDIUM(jt)              &
               , quantity  = I_QUANTITY(jt)         &
               , cask_i = CASK_I(jt,:)              &
               , cask_r = CASK_R(jt,:)              &
               , cask_s = CASK_S(jt,:)              &
               )
       END DO set_loop

    END DO tracer_loop

    CALL end_message_si(modstr, 'TRACER REQUEST', substr)

  END SUBROUTINE ptrac_new_tracer
! ------------------------------------------------------------------------

! ------------------------------------------------------------------------
  SUBROUTINE ptrac_init_memory

#ifdef MESSYCHANNELS
    ! BMIL
    USE messy_main_mpi_si,     ONLY: p_parallel_io
    USE messy_main_data_si,    ONLY: nproma, ngpblks

    ! ECHAM5
    USE mo_memory_base,    ONLY: new_stream                   &
                               , add_stream_element           &
                               , default_stream_setting       &
                               , t_stream, ARRAY1D
    ! SMCL
    USE messy_main_tracer,     ONLY: I_AEROSOL_MODE, R_AEROSOL_DENSITY

    IMPLICIT NONE
    INTRINSIC :: ANY, MAXVAL, TRIM

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = ' ptrac_init_memory'
    TYPE (t_stream), POINTER    :: stream_ptr => NULL()
    INTEGER                     :: jt
    LOGICAL                     :: lstream
    !
    ! NUMBER OF MODES
    INTEGER                     :: maxmode, jm

    ! STREAM REQUIRED ?
    lstream = (ANY(I_MEDIUM(:) == AEROSOL))
    IF (.NOT. lstream) RETURN

    CALL start_message_si(modstr, 'STREAM DEFINITION', substr)

    ! define new stream for ptrac
    CALL new_stream (stream_ptr,modstr, lrerun=.false., lpost=.false.)
    CALL default_stream_setting(stream_ptr &
         ,lpost =.FALSE.          &
         ,lrerun = .false.        &
         ,lav = .false.           &
         ,lsd = .false.           &
         ,contnorest=.true.       & ! mz_pj_20050805
         )

    ! INFORMATION IN STREAM
    aerosol_cond: IF (ANY(I_MEDIUM(:) == AEROSOL)) THEN
       maxmode = MAXVAL(CASK_I(1:I_NTRAC,I_AEROSOL_MODE))
       !
       IF (p_parallel_io) WRITE(*,*) ' ... wetradius'
       CALL add_stream_element(stream_ptr &
            , 'wetradius'                 &
            , radius_ptr                  &
            , ktrac = maxmode             &
            , units ='m'                  &
            , longname='const. ambient aerosol radius' &
            , lrerun = .FALSE.            &
            , lpost = .FALSE.             &
            , lav = .FALSE.               &
            , lsd = .FALSE.               &
            )
       
       IF (p_parallel_io) WRITE(*,*) ' ... density'
       CALL add_stream_element(stream_ptr &
            , 'densaer'                   &
            , density_ptr                 &
            , ktrac = maxmode             &
            , units ='kg/m3'              &
            , longname='aerosol density'  &
            , lrerun = .FALSE.            &
            , lpost = .FALSE.             &
            , lav = .FALSE.               &
            , lsd = .FALSE.               &
            )
       
       IF (p_parallel_io) WRITE(*,*) ' ... sigma'
       CALL add_stream_element (stream_ptr    &
            , 'sigma'                         &
            , sigma_ptr                       &
            , units=' '                       &
            , nsize= maxmode                  &
            , repr=ARRAY1D                    &
            , longname='standard deviation of aerosol mode' &
            , lrerun=.FALSE.                  &
            , lpost=.FALSE.                   &
            , lav = .FALSE.                   &
            , lsd = .FALSE.                   &
            )
       
       ! INITIALIZE RADIUS AND SIGMA
       ! -> FIRST TRACER OF MODE DETERMINES RADIUS, SIGMA, and DENSITY
       sigma_ptr(:) = 1.59_dp ! DEFAULT
       DO jm = 1, maxmode
          DO jt=1, I_NTRAC
          IF (CASK_I(jt,I_AEROSOL_MODE) == jm) THEN
            radius_ptr(:,:,jm,:) = R_AEROSOL_RADIUS(jt)
            sigma_ptr(jm) = R_AEROSOL_SIGMA(jt)
            density_ptr(:,:,jm,:) = (2._dp/3._dp) * &
                 CASK_R(jt, R_AEROSOL_DENSITY)
            EXIT
          ENDIF
          END DO
       END DO

    END IF aerosol_cond

    CALL end_message_si(modstr, 'STREAM DEFINITION', substr)
#endif

  END SUBROUTINE ptrac_init_memory
! ------------------------------------------------------------------------

! ------------------------------------------------------------------------
  SUBROUTINE ptrac_init_tracer

#ifdef ECHAM5
    ! BMIL
    USE messy_main_tracer_si, ONLY: tracer_init

    IMPLICIT NONE

    CALL tracer_init(modstr)
#endif

  END SUBROUTINE ptrac_init_tracer
! ------------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE ptrac_read_nml_cpl(status, iou)

    ! SMCL
    USE messy_main_tools,         ONLY: read_nml_open, read_nml_check &
                                      , read_nml_close
    USE messy_main_tracer,        ONLY: get_tracer_set, t_trinfo_list &
                                      , param2string &
                                      , MAX_CASK_I, NAMES_CASK_I &
                                      , MAX_CASK_R, NAMES_CASK_R &
                                      , MAX_CASK_S, NAMES_CASK_S
    USE messy_main_constants_mem, ONLY: STRLEN_MEDIUM


    IMPLICIT NONE
    INTRINSIC :: TRIM

    ! I/O
    INTEGER, INTENT(OUT) :: status     ! error status
    INTEGER, INTENT(IN)  :: iou        ! I/O unit

    ! (LOCAL) NAMELIST VARIABLES
    CHARACTER(LEN=*), PARAMETER :: substr = 'ptrac_read_nml_cpl'

    NAMELIST /CPL/ C_SETS, C_NAME, C_UNIT            &
                 , I_TYPE, I_MEDIUM, I_QUANTITY      &
                 , CASK_I, CASK_S, CASK_R            &
                 , R_AEROSOL_SIGMA, R_AEROSOL_RADIUS


    ! LOCAL
    LOGICAL :: lex      ! file exists ?
    INTEGER :: fstat    ! file status
    INTEGER :: jt, j
    CHARACTER(LEN=STRLEN_MEDIUM) :: attstr

    status = 1

    CALL read_nml_open(lex, substr, iou, 'CPL', modstr)
    IF (.not.lex) RETURN    ! <modstr>.nml does not exist

    READ(iou, NML=CPL, IOSTAT=fstat)
    CALL read_nml_check(fstat, substr, iou, 'CPL', modstr)
    IF (fstat /= 0) RETURN  ! error while reading namelist

    ! DIAGNOSE NAMELIST AND SET GLOBAL SWITCHES

    tracer_loop: DO jt = 1, NMAXNTRAC

       IF (TRIM(C_NAME(jt)%name) == '') CYCLE

       WRITE(*,*) '.........................................................'
       WRITE(*,*) 'TRACER NO. ',jt
       IF (TRIM(C_NAME(jt)%subname) ==  '') THEN
          WRITE(*,*) '  NAME         = ', TRIM(C_NAME(jt)%name)
       ELSE
          WRITE(*,*) '  NAME         = ', &
               TRIM(C_NAME(jt)%name)//'_'//TRIM(C_NAME(jt)%subname)
       END IF

       WRITE(*,*) '  UNIT         = ', TRIM(C_UNIT(jt))

       attstr = param2string(I_TYPE(jt), 'type')
       WRITE(*,*) '  TYPE         = '//TRIM(attstr)

       attstr = param2string(I_MEDIUM(jt), 'medium')
       WRITE(*,*) '  MEDIUM       = '//TRIM(attstr)

       attstr = param2string(I_QUANTITY(jt), 'quantity')
       WRITE(*,*) '  QUANTITY     = '//TRIM(attstr)
       
       DO j=1, MAX_CASK_I
          attstr = param2string(CASK_I(jt,j),'switch')
          WRITE(*,*) NAMES_CASK_I(j)//': ', TRIM(attstr) 
       END DO

       DO j=1, MAX_CASK_R
          WRITE(*,*) NAMES_CASK_R(j)//': ',CASK_R(jt,j)
       END DO

       DO j=1, MAX_CASK_S
          WRITE(*,*) NAMES_CASK_S(j)//': '//TRIM(CASK_S(jt,j))
       END DO

    END DO tracer_loop

    CALL read_nml_close(substr, iou, modstr)

    status = 0  ! no ERROR

  END SUBROUTINE ptrac_read_nml_cpl
! ----------------------------------------------------------------------

! **********************************************************************
END MODULE messy_ptrac_si
! **********************************************************************
